@@ -48,7 +48,7 @@
 #include "network.h"
 #include "var.h"
 #include "main.h"
-#include "db.h"
+#include "mydb.h"
 #include "dc_manage.h"
 #include "he3.h"
 #include "user_manage.h"
@@ -127,60 +127,6 @@ static MY_DIR decode_direction(GString *input, unsigned int *level)
 		return MY_DIR_ERROR;
 }
 
-#if 0
-/***********************************/
-/* compute and send xfer direction */
-/**********************************************************/
-/* output: MY_DIR_ERROR, MY_DIR_DOWNLOAD or MY_DIR_UPLOAD */
-/**********************************************************/
-static MY_DIR send_direction(int sck, GString *input)
-{
-	char *t;
-	int a;
-	GString *sent;
-	int ret;
-
-	if(input==NULL)
-		return MY_DIR_ERROR;
-
-	if(strncmp(input->str,"$Direction ",sizeof("$Direction ")-1))
-		return MY_DIR_ERROR;
-
-	t=input->str+ sizeof("$Direction ")-1;
-
-	if(!strncmp(t,"Download",sizeof("Download")-1))
-	{
-		ret=MY_DIR_UPLOAD;
-		sent=g_string_new("$Direction Upload ");
-	}
-	else if(!strncmp(t,"Upload",sizeof("Upload")-1))
-	{
-		ret=MY_DIR_DOWNLOAD;
-		sent=g_string_new("$Direction Download ");
-	}
-	else
-		return MY_DIR_ERROR;
-		
-	/* add a number after the string */
-	g_string_sprintfa(sent,"%u|",rand()%9998+1);
-
-	disp_msg(DEBUG_MSG,"send_direction",sent->str,NULL);
-
-#if 0
-	a=write(sck,sent->str,sent->len);			/* @@@ replaced by send */
-	a=(a!=sent->len);
-#else
-	a=( (send(sck,sent->str,sent->len,MSG_NOSIGNAL)) != sent->len );
-#endif
-	g_string_free(sent,TRUE);
-
-	if(a)
-		return MY_DIR_ERROR;
-	else
-		return ret;
-}
-#endif
-
 /***************************************************************************/
 /* create a lock inside the 'created_lock' array and send it on the socket */
 /***************************************************************************/
@@ -199,13 +145,6 @@ static int send_a_lock(int sck,char *created_lock)
 	created_lock[i++]='|';					/* and the | at the end */
 	created_lock[i++]='\0';
 
-#if 0
-	if(write(sck,"$Lock ",6)!=6)															/* @@@ replaced by following code */
-		return 1;		/* fails */
-
-	if(write(sck,created_lock,strlen(created_lock))!=strlen(created_lock))	/* @@@ replaced by following code */
-		return 1;		/* fails */
-#else
 	{
 		gchar *v;
 		int ret;
@@ -217,7 +156,6 @@ static int send_a_lock(int sck,char *created_lock)
 		if(ret)
 			return 1;	/* fails */
 	}
-#endif
 
 	created_lock[ln]='\0';		/* truncate to keep only the useful part */
 	return 0;		/* ok */
@@ -230,7 +168,7 @@ static int send_a_lock(int sck,char *created_lock)
 /* output: 0=ok, 1=error                         */
 /*         if ok, *file_len is the reported size */
 /*************************************************/
-static int send_file_length(int sck,char *filename, int val, unsigned long *file_len)
+static int send_file_length(int sck,char *filename, unsigned long *file_len)
 {
 	GString *out;
 	int res;
@@ -246,11 +184,7 @@ static int send_file_length(int sck,char *filename, int val, unsigned long *file
 
 	printf("%s\n%s\n",filename,out->str);
 
-#if 0
-	res=(write(sck,out->str,out->len)!=out->len);						/* @@@ replaced by following code */
-#else
 	res=(send(sck,out->str,out->len,MSG_NOSIGNAL)!=out->len);
-#endif
 	g_string_free(out,TRUE);
 	return res;
 }
@@ -293,7 +227,7 @@ static int send_file_data(int sck,char *filename, int start_pos, unsigned long f
 
 		act->last_touch=time(NULL);
 
-		get_slices(bl_semid,sizeof(buf)/512);							/* obtain upload authorization */
+		get_ul_slices(bl_semid,sizeof(buf)/512);							/* obtain upload authorization */
 		res=send(sck,buf,sizeof(buf),MSG_NOSIGNAL /* |MSG_WAITALL */ );
 
 		act->last_touch=time(NULL);
@@ -316,7 +250,7 @@ static int send_file_data(int sck,char *filename, int start_pos, unsigned long f
 		
 		act->last_touch=time(NULL);
 
-		get_slices(bl_semid,(remain+511)/512);							/* obtain upload authorization */
+		get_ul_slices(bl_semid,(remain+511)/512);							/* obtain upload authorization */
 		res=send(sck,buf,remain,MSG_NOSIGNAL /* |MSG_WAITALL */ );
 
 		act->last_touch=time(NULL);
@@ -350,7 +284,7 @@ static int send_array_data(int sck,GByteArray *ba,WAIT_ACT *act)
 	{
 		act->last_touch=time(NULL);
 
-		get_slices(bl_semid,BLOCK_SIZE/512);							/* obtain upload authorization */
+		get_ul_slices(bl_semid,BLOCK_SIZE/512);							/* obtain upload authorization */
 		res=send(sck,ba->data+cur_pos,BLOCK_SIZE,MSG_NOSIGNAL /* |MSG_WAITALL */ );
 
 		act->last_touch=time(NULL);
@@ -368,7 +302,7 @@ static int send_array_data(int sck,GByteArray *ba,WAIT_ACT *act)
 	{
 		act->last_touch=time(NULL);
 
-		get_slices(bl_semid,(remain+511)/512);							/* obtain upload authorization */
+		get_ul_slices(bl_semid,(remain+511)/512);							/* obtain upload authorization */
 		res=send(sck,ba->data+cur_pos,remain,MSG_NOSIGNAL /* |MSG_WAITALL */ );
 
 		act->last_touch=time(NULL);
@@ -398,8 +332,6 @@ static void unconvert_path(char *str)
 /* a $Send| which is processed by this function    */
 /**************************************************************************/
 /* this function should return the list of all files shared by the client */
-/* This version doesn't support this feature and instead report a dummy   */
-/* list size and always says "no more slot".                              */
 /**************************************************************************/
 static int com_up_get_list_len_process(const char *cmd,WAIT_ACT *act,int sck,GString *input, char *xtra_param)
 {
@@ -560,6 +492,7 @@ static int com_up_get_process(const char *cmd,WAIT_ACT *act,int sck,GString *inp
 	int val;
 	GString *str2;
 	unsigned long file_len;
+	int virtual;
 
 	t=input->str+strlen(cmd);
 	SKIP_SPACE(t)
@@ -593,7 +526,7 @@ static int com_up_get_process(const char *cmd,WAIT_ACT *act,int sck,GString *inp
 
 	unconvert_path(fullpathname);
 
-	if(!file_in_db(fullpathname))		/* is it a shared file ? */
+	if(!file_in_db(fullpathname,&virtual))		/* is it a shared file ? */
 	{
 		GString *err_msg;
 
@@ -605,18 +538,32 @@ static int com_up_get_process(const char *cmd,WAIT_ACT *act,int sck,GString *inp
 		return 1;
 	}
 
-	if(((unsigned char)fullpathname[0])==255)
+	if(hide_absolute==0)
+	{
+		if(((unsigned char)fullpathname[0])==255)
+			fullpathname[0]='/';
+	}
+	else
+	{
+		/* we always prepend a '/' at the beginning of the fullpathname */
+		/* this is done in a very dirty way. Due to the fact that the command is always before the fullname */
+		/* and we won't use it anymore, we "accidentally" trash it :) */
+		fullpathname--;
 		fullpathname[0]='/';
+	}
 
 	/* it is time to verify if a slot is available */
 	LOCK_WRITE(user_info);
-	if((dl_on==0)||(free_dl_slot<1))
+	if((dl_on==0)||(free_dl_slot<1)||(virtual!=0))
 	{
-		/* no free slot or no download allowed */
-		UNLOCK_WRITE(user_info);
-
-		send_dc_line(sck,"$MaxedOut",NULL);
-		return 1;	/* abort */
+		if(!user_has_flag(act->remote_nick->str,"IGNORE_SLOT_LIMIT"))
+		{
+			/* no free slot or no download allowed */
+			UNLOCK_WRITE(user_info);
+	
+			send_dc_line(sck,"$MaxedOut",NULL);
+			return 1;	/* abort */
+		}
 	}
 
 	/* lock a slot */
@@ -626,7 +573,7 @@ static int com_up_get_process(const char *cmd,WAIT_ACT *act,int sck,GString *inp
 	send_dc_line(sck,"$Capabilities","SEND_WITH_SIZE",NULL);
 #endif
 
-	if(send_file_length(sck,fullpathname,val,&file_len))
+	if(send_file_length(sck,fullpathname,&file_len))
 	{
 		disp_msg(DEBUG_MSG,"com_up_get_process","fail2",NULL);
 		LOCK_WRITE(user_info);
@@ -645,9 +592,34 @@ static int com_up_get_process(const char *cmd,WAIT_ACT *act,int sck,GString *inp
 		return 1;			/* abort */
 	}
 
-	
 	disp_msg(DEBUG_MSG,"com_up_get_process","str2:",str2->str,NULL);
 	
+	if(!strncmp(str2->str,"$ChgSSP ",strlen("$ChgSSP ")))
+	{
+		val=atoi(str2->str+strlen("$ChgSSP "));	/* new start position */
+		
+		{	/* Send the reply */
+			GString *start_pos_msg;
+
+			start_pos_msg=g_string_new("$StartPos ");
+			g_string_sprintfa(start_pos_msg,"%d",val);
+			send_dc_line(sck,start_pos_msg->str,NULL); /* Send the reply */
+			g_string_free(start_pos_msg,TRUE);
+		}
+
+		--val; /* new start position minus 1 */		 
+		
+		str2=get_a_dc_line(sck);
+		if(str2==NULL)
+		{
+			disp_msg(DEBUG_MSG,"com_up_get_process","fail4",NULL);
+			LOCK_WRITE(user_info);
+			free_dl_slot++;
+			UNLOCK_WRITE(user_info);
+			return 1;			/* abort */
+		}
+	}
+
 	if(strcmp(str2->str,"$Send|"))
 	{
 		if(strncmp(str2->str,"$Send ",strlen("$Send ")))
@@ -725,6 +697,7 @@ static int com_up_getfilelength_process(const char *cmd,WAIT_ACT *act,int sck,GS
 	char *t;
 	char *fullpathname;
 	unsigned long file_len;
+	int virtual;
 
 	t=input->str+strlen(cmd);
 	SKIP_SPACE(t)
@@ -743,7 +716,7 @@ static int com_up_getfilelength_process(const char *cmd,WAIT_ACT *act,int sck,GS
 
 	unconvert_path(fullpathname);
 
-	if(!file_in_db(fullpathname))		/* is it a shared file ? */
+	if(!file_in_db(fullpathname,&virtual))		/* is it a shared file ? */
 	{
 		GString *err_msg;
 
@@ -755,10 +728,22 @@ static int com_up_getfilelength_process(const char *cmd,WAIT_ACT *act,int sck,GS
 		return 1;
 	}
 
-	if(((unsigned char)fullpathname[0])==255)
+	if(hide_absolute==0)
+	{
+		if(((unsigned char)fullpathname[0])==255)
+			fullpathname[0]='/';
+	}
+	else
+	{
+		/* we always prepend a '/' at the beginning of the fullpathname */
+		/* this is done in a very dirty way. Due to the fact that the command is always before the fullname */
+		/* and we won't use it anymore, we "accidentally" trash it :) */
+		fullpathname--;
 		fullpathname[0]='/';
+	}
 
-	if(send_file_length(sck,fullpathname,0,&file_len))
+
+	if(send_file_length(sck,fullpathname,&file_len))
 	{
 		disp_msg(DEBUG_MSG,"com_up_getfilelength_process","fail2",NULL);
 		return 1;			/* abort */
@@ -883,6 +868,8 @@ static int copie_fd_to_file(int remote, FILE *local, unsigned long amount,WAIT_A
 
 		nb=min(amount,sizeof(buf));
 
+		get_dl_slices(bl_semid,(nb+1023)/1024);
+
 		/* touch the action slot to avoid timeout */
 		act->last_touch=time(NULL);
 		ret=recv(remote,buf,nb,MSG_WAITALL|MSG_NOSIGNAL);
@@ -1217,8 +1204,11 @@ static int start_a_xdownload(WAIT_ACT *act, WAIT_REVCON *act_to_do)
 	GString *get_str;
 	GString *lfile;
 	GString *input=NULL;
-	unsigned long dummy;
+	unsigned long remote_file_size;
+	unsigned long start_position;
 	FILE *local_file=NULL;
+	int no_link=1;
+	int have_worked=0;
 
 	disp_msg(DEBUG_MSG,"start_a_xdownload",act_to_do->action_to_do->str,NULL);
 	G_LOCK(waiting_action);
@@ -1231,7 +1221,7 @@ static int start_a_xdownload(WAIT_ACT *act, WAIT_REVCON *act_to_do)
 	gdl_id=strtoul(fields[1],NULL,10);
 	/* fields[2] is the remote nick */
 
-	if(do_gdl_start(gdl_id, fields[2], act->thread_id, &get_str,&lfile)==0)
+	if(do_gdl_start(gdl_id, fields[2], act->thread_id, &get_str,&lfile,&start_position)==0)
 	{
 		int is_fatal=0;
 		unsigned long total_pos=0;
@@ -1267,7 +1257,7 @@ static int start_a_xdownload(WAIT_ACT *act, WAIT_REVCON *act_to_do)
 			goto end_on_error;
 		}
 
-		if(extract_filelength(input->str,&dummy))
+		if(extract_filelength(input->str,&remote_file_size))
 			goto end_on_error;
 
 		local_file=fopen(lfile->str,"wb");
@@ -1297,6 +1287,8 @@ static int start_a_xdownload(WAIT_ACT *act, WAIT_REVCON *act_to_do)
 			if(amount==0)
 				break;
 
+			get_dl_slices(bl_semid,(amount+1023)/1024);
+
 			act->last_touch=time(NULL);
 			ret=recv(act->sock_fd,buf,amount,MSG_WAITALL|MSG_NOSIGNAL);
 			if((ret==-1)||(ret==0))
@@ -1322,11 +1314,24 @@ static int start_a_xdownload(WAIT_ACT *act, WAIT_REVCON *act_to_do)
 		}
 
 		do_gdl_success(gdl_id, act->thread_id,1);		/* currently, it is not yet possible to link multiple downloads on the same link */
+
+		/* it is possible to link task in 1 case. If our download end position is the remote file size */
+		/* because in this case, the remote client expects a client. In all other cases, it still sends data */
+		if((start_position+total_pos)==remote_file_size)
+			no_link=0;
+
+		have_worked=1;
 	}
 	leave:
 
 	if(input!=NULL)
 		g_string_free(input,TRUE);
+
+	if(have_worked)
+	{
+		gdl_wake_up_sources(fields[2],0);
+	}
+
 	g_strfreev(fields);
 	G_LOCK(waiting_action);
 	if(act->disp_info)
@@ -1336,7 +1341,7 @@ static int start_a_xdownload(WAIT_ACT *act, WAIT_REVCON *act_to_do)
 	}
 	act->run_task=NULL;						/* dissociate the task from this thread */
 	G_UNLOCK(waiting_action);
-	return 1;		/* never link commands */
+	return no_link;		/* never link commands (except in few cases) */
 }
 
 #if 0
@@ -1451,6 +1456,7 @@ static void generate_shared_file_list(WAIT_REVCON *act_to_do, GString *content)
 
 	content=my_g_string_dos2unix(content);
 
+	disp_msg(LS_BEGIN,"",act_to_do->remote_nick->str,NULL);
 	dir_part=g_string_new("");
 	temp_str=g_string_new("");
 	
@@ -1682,7 +1688,7 @@ static void send_client_capabilities(WAIT_ACT *act)
 #if 0
 	str=g_string_new("SEND_WITH_SIZE");
 #else
-	str=g_string_new("GetFileLength");
+	str=g_string_new("GetFileLength$ChgSSP");
 #endif
 
 	send_dc_line(act->sock_fd,"$Capabilities",str->str,NULL);
@@ -1696,6 +1702,7 @@ void requeue_act_to_do(WAIT_REVCON *act_to_do)
 {
 	if(act_to_do!=NULL)
 	{	
+#if 0
 		/* if the error can be recovered, we will try to restart the command a bit later */
 		if(!strncmp(act_to_do->action_to_do->str,"DL/",3))
 		{
@@ -1740,7 +1747,9 @@ void requeue_act_to_do(WAIT_REVCON *act_to_do)
 			g_string_free(local,TRUE);
 			g_string_free(org_remote,TRUE);
 		}
-		else if(!strncmp(act_to_do->action_to_do->str,"LS/",3))
+		else 
+#endif
+		if(!strncmp(act_to_do->action_to_do->str,"LS/",3))
 		{
 			GString *sim;
 			sim=g_string_new("");
@@ -1755,7 +1764,20 @@ void requeue_act_to_do(WAIT_REVCON *act_to_do)
 		}
 		else if(!strncmp(act_to_do->action_to_do->str,"XDL",3))
 		{
-			disp_msg(INFO_MSG,"requeue_act_to_do","XDL requeue",NULL);
+			gchar **fields;
+			disp_msg(DEBUG_MSG,"requeue_act_to_do","XDL requeue",NULL);
+			
+			fields=g_strsplit(act_to_do->action_to_do->str,"|",0);
+			if((fields[0]!=NULL)&&(fields[1]!=NULL)&&(fields[2]!=NULL))
+			{
+#if 0
+				/* fail a GDL entry for the given nickname (no filename provided) */
+				do_gdl_fail(strtoul(fields[1],NULL,10),fields[2],NULL,0);
+#else
+				do_gdl_abort(strtoul(fields[1],NULL,10),fields[2]);
+#endif
+			}
+			g_strfreev(fields);
 		}
 	}
 }
@@ -1826,6 +1848,14 @@ static void *connect_me_thread(WAIT_ACT *act)
 	/* My nick is here */
 	input=g_string_erase(input,0,8);		/* remove the "$MyNick " at the beginning */
 	input=g_string_truncate(input,input->len-1);		/* remove the trailing | */
+
+	if(user_has_flag(input->str,"IGNORE_XFER"))
+	{
+		g_string_free(input,TRUE);
+		input=NULL;
+		goto thread_over;
+	}
+
 	G_LOCK(waiting_action);
 	act->remote_nick=input;							/* set the remote nick */
 	G_UNLOCK(waiting_action);
@@ -1910,7 +1940,7 @@ static void *connect_me_thread(WAIT_ACT *act)
 		if((our_dir==MY_DIR_UPLOAD)&&(remote_dir==MY_DIR_UPLOAD))
 		{
 			/* nobody wants to download, ends the connection */
-			disp_msg(INFO_MSG,"connect_me_thread","Null download connection",NULL);
+			disp_msg(INFO_MSG,"connect_me_thread","Null download connection",act->remote_nick->str,NULL);
 
 			/* if I am MY_DIR_UPLOAD, act_to_do==NULL */
 			g_string_free(input,TRUE);
@@ -1925,14 +1955,22 @@ static void *connect_me_thread(WAIT_ACT *act)
 
 			if(remote_level>my_level)
 			{
-				/* the remote side wins, I switch to upload mode */
-				if(act_to_do!=NULL)
+				if(with_dl_force==0)
 				{
-					prepend_action_to_do(act_to_do);			/* reput the action at the beginning of the waiting action */
-					act_to_do=NULL;
+					/* the remote side wins, I switch to upload mode */
+					/* (but only if download forcing is not enabled) */
+					if(act_to_do!=NULL)
+					{
+						prepend_action_to_do(act_to_do);			/* reput the action at the beginning of the waiting action */
+						act_to_do=NULL;
+					}
+					our_dir=MY_DIR_UPLOAD;
+					disp_msg(INFO_MSG,"connect_me_thread","twin download connection","low priority:switching to upload",NULL);
+				}
+				else
+				{
+					disp_msg(INFO_MSG,"connect_me_thread","twin download connection","low priority:direction override",NULL);
 				}
-				our_dir=MY_DIR_UPLOAD;
-				disp_msg(INFO_MSG,"connect_me_thread","twin download connection","low priority:switching to upload",NULL);
 			}
 		}
 
@@ -1941,7 +1979,7 @@ static void *connect_me_thread(WAIT_ACT *act)
 		/* now, reply to the first $Lock. It seems strange but we must do this in this order or it doesn't work */
 		if((i=do_unlock_access(act->sock_fd,input))!=0)
 		{
-			disp_msg(ERR_MSG,"sub_lock_process","|lu",(unsigned long)act->thread_id,"unable to unlock","|d",(int)i,NULL);
+			disp_msg(ERR_MSG,"sub_lock_process","|lu",(unsigned long)act->thread_id,"unable to unlock","|d",(int)i,"nick","|s",act->remote_nick->str,NULL);
 			goto requeue_and_thread_over;
 		}
 
@@ -1980,9 +2018,11 @@ static void *connect_me_thread(WAIT_ACT *act)
 		/* we have received the $Key string. It is the end of the handshake */
 
 		disp_msg(DEBUG_MSG,"connect_me_thread",input->str,NULL);
-		if(verify_key(c_lock,input))
+
+		/* allow lazy key check if flag allows it and if it is a download only */
+		if(verify_key(c_lock,input,  (((our_dir==MY_DIR_DOWNLOAD) && (with_lazy_key_check))?1:0) ))
 		{
-			disp_msg(ERR_MSG,"connect_me_thread","|lu",(unsigned long)act->thread_id,"Key invalid",NULL);
+			disp_msg(ERR_MSG,"connect_me_thread","|lu",(unsigned long)act->thread_id,"Key invalid","nick","|s",act->remote_nick->str,NULL);
 			goto requeue_and_thread_over;
 		}
 		g_string_free(input,TRUE);				/* no more useful */
@@ -2064,8 +2104,8 @@ static void *do_connect_to_me(GString *host_port)
 	{
 		delete_uaddr_entry_by_name(hst);			/* fail to open remote address, invalid the uaddr entry to avoid infinite tries */
 															/* with an invalid address */
-		free(hst);
 		disp_msg(ERR_MSG,"do_connect_to_me","unable to contact",hst,NULL);
+		free(hst);
 		return NULL;
 	}
 	free(hst);
@@ -2080,7 +2120,7 @@ static void *do_connect_to_me(GString *host_port)
 		return NULL;
 	}
 
-	disp_msg(ERR_MSG,"do_connect_to_me","creating thread info","|lu",(unsigned long)nw_act,NULL);
+	disp_msg(DEBUG_MSG,"do_connect_to_me","creating thread info","|lu",(unsigned long)nw_act,NULL);
 	nw_act->running=3;			/* thread init */
 	nw_act->cnx_type=CNX_ON_CONNECT;
 	nw_act->last_touch=time(NULL);
@@ -2138,6 +2178,15 @@ int connecttome_process(const char *cmd,int sck,GString *input, char *xtra_param
 	GString *dest;
 	pthread_t num;
 
+	/* why wasting time and bandwidth to do something which cannot succeed */
+	/* if we are in active mode, we receive a connect to me only if someone */
+	/* in active mode wants to download from us. If no download slot is available */
+	/* we can simply ignore the query */
+	if((!behind_fw)&&(free_dl_slot==0))
+	{
+		return 1;
+	}
+
 	t=input->str+strlen(cmd);
 	SKIP_SPACE(t)
 	
@@ -2244,6 +2293,7 @@ static int to_process(const char *cmd,int sck,GString *input, char *xtra_param)
 	char *t;
 	char *nic;
 	char *msg;
+	GString *org_msg=NULL;
 
 	t=input->str+strlen(cmd);
 	SKIP_SPACE(t)
@@ -2252,9 +2302,15 @@ static int to_process(const char *cmd,int sck,GString *input, char *xtra_param)
 	{
 		bad_param:
 		disp_msg(ERR_MSG,"to_process","bad param",NULL);
+		if(org_msg!=NULL)
+			g_string_free(org_msg,TRUE);
 		return 1;
 	}
 	
+	org_msg=g_string_new(input->str);
+	if(org_msg==NULL)
+		goto bad_param;
+
 	/* skip my nickname */
 	t=strchr(t,' ');
 	if(t==NULL)
@@ -2274,15 +2330,37 @@ static int to_process(const char *cmd,int sck,GString *input, char *xtra_param)
 		goto bad_param;
 
 	*t++='\0';
+	if(user_has_flag(nic,"IGNORE_PMSG"))		/* ignore private message from this user */
+		return 1;
+
 	t++;
 	msg=t;
 	strrchr(t,'|')[0]='\0';		/* can never fail */
 
-	
 	subst_char(msg,'\n','\r');
 
 	disp_msg(PRIV_MSG,NULL,nic,msg,NULL);
-	
+
+	/* and relay the message to all client */
+	/* but only if it is not a relayed message */
+	if(input->str[0]=='$')
+	{
+		org_msg=g_string_prepend_c(org_msg,'*');
+
+		/* search the end of the sender nickname in the message */
+		t=strstr(org_msg->str+(msg-input->str),"> ");
+		if((t!=NULL)&&(hubname!=NULL)&&(hubname->len!=0))
+		{
+			gint pos=t-org_msg->str;
+			org_msg=g_string_insert(org_msg,pos," ()");
+			pos+=2;
+			org_msg=g_string_insert(org_msg,pos,hubname->str);	/* put the hubname between the 2 parenthesis */
+		}
+
+		send_dc_line_to_dctc_link_direct(NULL,org_msg->str);
+	}
+
+	g_string_free(org_msg,TRUE);
 	return 1;		/* end */
 }
 
@@ -2387,10 +2465,24 @@ static int hello_process(const char *cmd,int sck,GString *input, char *xtra_para
 			UNLOCK_READ(user_info);
 		}
 		add_user_to_user_list(t);		/* USER_IN_MSG */
+
+		if((with_incoming_wake_up)&&(!tos_entry_exists(WAKEUGDL_TOSKEY,nickname,strlen(nickname),0)))
+		{	/* if option is enabled and this user waiting downloads have not been waked up since enough time, wake up them */
+			add_tos_entry_v1_uniq(WAKEUGDL_TOSKEY,min_gdl_wake_up_delay,nickname,strlen(nickname),NULL,0);
+			send_dc_line_to_dctc_link(NULL,sck,"/WAKEUGDL",nickname,NULL);
+			gdl_wake_up_sources(nickname,0);
+		}
 	}
 	else if(msg_type==USER_OUT_MSG)
+	{
 		remove_user_to_user_list(t);	/* USER_OUT_MSG */
 
+		if(abort_upload_when_user_leaves)
+		{
+			terminate_upload_of_user(t);
+		}
+	}
+
 	return 1;		/* end */
 }
 
@@ -2700,29 +2792,32 @@ static int search_process(const char *cmd,int sck,GString *input, char *xtra_par
 		set_xtra_information_cnx(reply_to+4,PASSIVE,NULL);
 
 		delete_uaddr_entry_by_name(reply_to+4);			/* user is not a valid direct download because he is passive */
+		if(!user_has_flag(reply_to+4,"IGNORE_SRCH"))
+		{	/* user query not to ignore ? */
 #if 0
-		/* currently, we reply to everybody */
-		if(!behind_fw)
+			/* currently, we reply to everybody */
+			if(!behind_fw)
 #endif
-		{
-			/* don't reply if we are also behind a firewall */
-			/* because no transfer is possible */
-			LOCK_READ(user_info);
-			if(strcmp(reply_to+4,nickname))
-			{	/* don't reply to our own query */
-				UNLOCK_READ(user_info);
-	
-				search_in_db(sck, reply_to+4,			/* don't put the hub: */
-										atoi(d),				/* wanted type */
-										(bool_a==0)?SA_NO_MATTER: ( (bool_b==0)? SA_AT_LEAST:SA_AT_MOST ),	/* size type */
-										strtoul(c,NULL,10),			/* wanted size */
-										e,
-										md5sum,				/* md5sum */
-										NULL);
-			}
-			else
 			{
-				UNLOCK_READ(user_info);
+				/* don't reply if we are also behind a firewall */
+				/* because no transfer is possible */
+				LOCK_READ(user_info);
+				if(strcmp(reply_to+4,nickname))
+				{	/* don't reply to our own query */
+					UNLOCK_READ(user_info);
+		
+					search_in_db(sck, reply_to+4,			/* don't put the hub: */
+											atoi(d),				/* wanted type */
+											(bool_a==0)?SA_NO_MATTER: ( (bool_b==0)? SA_AT_LEAST:SA_AT_MOST ),	/* size type */
+											strtoul(c,NULL,10),			/* wanted size */
+											e,
+											md5sum,				/* md5sum */
+											NULL);
+				}
+				else
+				{
+					UNLOCK_READ(user_info);
+				}
 			}
 		}
 	}
@@ -2768,6 +2863,7 @@ static int sresult_process(const char *cmd,int sck,GString *input, char *xtra_pa
 	char *t;
 	char *nic, *fic, *ficsize, *ratio, *hbname, *hbip;
 	char *u;
+	int res;
 
 	t=input->str+strlen(cmd);
 	SKIP_SPACE(t)
@@ -2789,6 +2885,15 @@ static int sresult_process(const char *cmd,int sck,GString *input, char *xtra_pa
 		goto bad_param;
 	*t++='\0';
 
+	LOCK_READ(user_info);
+	res=strcmp(nic,nickname);
+	UNLOCK_READ(user_info);
+	if(res==0)
+		return 1;				/* it is our nickname */
+
+	if(user_has_flag(nic,"IGNORE_SREP"))
+		return 1;				/* ignore this user replies */
+
 	/* get filename */
 	fic=t;
 	t=strchr(fic,5);
@@ -2845,6 +2950,20 @@ static int sresult_process(const char *cmd,int sck,GString *input, char *xtra_pa
 			free(true_fname);
 	}
 
+	{
+		int ratio_free, ratio_ttl;
+
+		if(sscanf(ratio,"%d/%d",&ratio_free,&ratio_ttl)==2)
+		{
+			if((ratio_free>0)&&(with_sr_wake_up)&&(!tos_entry_exists(WAKEUGDL_TOSKEY,nic,strlen(nic),0)))
+			{	/* if there is free slot and option is enabled and this user waiting downloads have not been waked up since enough time, wake up them */
+				add_tos_entry_v1_uniq(WAKEUGDL_TOSKEY,min_gdl_wake_up_delay,nic,strlen(nic),NULL,0);
+				send_dc_line_to_dctc_link(NULL,sck,"/WAKEUGDL",nic,NULL);
+				gdl_wake_up_sources(nic,0);
+			}
+		}
+	}
+
 	return 1;		/* end */
 }
 
@@ -3036,6 +3155,7 @@ static CMD_REPLY cmd_reply[]=	{
 											{"$ConnectToMe "   , sizeof("$ConnectToMe ")-1   ,connecttome_process   ,NULL},
 											{"$RevConnectToMe ", sizeof("$RevConnectToMe ")-1,revconnecttome_process,NULL},
 											{"$To: "           , sizeof("$To: ")-1           ,to_process            ,NULL},
+											{"*To: "           , sizeof("*To: ")-1           ,to_process            ,NULL},	/* redirection support */
 											{"$ForceMove "     , sizeof("$ForceMove ")-1     ,force_process         ,NULL},
 											{"$GetPass"        , sizeof("$GetPass")-1        ,getpass_process       ,NULL},
 											{"$BadPass"        , sizeof("$BadPass")-1        ,badpass_process       ,NULL},
@@ -3049,6 +3169,49 @@ static CMD_REPLY cmd_reply[]=	{
 											{NULL,0,NULL},
 										};
 
+/***************************************************************************************/
+/* check if the given message is a kick/ban message and add the IP into the UADDR list */
+/***************************************************************************************/
+static void grab_ban_ip_fnc(GString *input)
+{
+	char *ip;
+	unsigned int p1,p2,p3,p4;
+
+	if(strncmp(input->str,"<Hub-Security> ",strlen("<Hub-Security> ")))
+		return;
+
+	ip=strstr(input->str," IP: ");
+	if(ip==NULL)
+		return;
+
+	ip+=strlen(" IP: ");
+
+	if(sscanf(ip,"%u.%u.%u.%u",&p1,&p2,&p3,&p4)==4)
+	{
+		char buf[512];
+
+		sprintf(buf,"%u.%u.%u.%u:412",p1,p2,p3,p4);
+		add_uaddr_entry_addr_dl_only(buf);
+	}
+}
+
+/*******************************************************/
+/* check if the given string has the following format: */
+/* <Hub-Security> The user xxx was kicked by yyy       */
+/*******************************************************/
+/* output: 1=yes, 0=no */
+/***********************/
+static int is_a_kick_message(char *str)
+{
+	if(strncmp(str,"<Hub-Security> The user ",strlen("<Hub-Security> The user ")))
+		return 0;
+
+	if(strstr(str+strlen("<Hub-Security> The user ")," was kicked by ")==NULL)
+		return 0;
+
+	return 1;
+}
+
 /*************************************************************************************/
 /* this function is called each time the client receives a full line from the server */
 /*****************************************************************************************/
@@ -3062,7 +3225,7 @@ int process_incoming_dc_data(int sck,GString *input)
 	if(input->len<2)		/* nothing to process */
 		return 1;			/* end */
 
-	if(input->str[0]=='$')
+	if((input->str[0]=='$')||(input->str[0]=='*'))
 	{	/* it is a command */
 		int i;
 
@@ -3093,7 +3256,13 @@ int process_incoming_dc_data(int sck,GString *input)
 			input->str[input->len-1]='\0';
 
 		subst_char(input->str,'\n','\r');
-		disp_msg(GLOB_CHAT_MSG,NULL,input->str,NULL);
+		if((hide_kick==0)||(!is_a_kick_message(input->str)))
+			disp_msg(GLOB_CHAT_MSG,NULL,input->str,NULL);
+
+		if(grab_ban_ip)
+		{
+			grab_ban_ip_fnc(input);
+		}
 		return 1;		/* end */
 	}
 }
