@@ -46,12 +46,19 @@
 #include "user_manage.h"
 #include "uaddr.h"
 #include "macro.h"
+#include "dc_com.h"
+#include "status.h"
+#include "sema.h"
+#include "timed_out_string.h"
+#include "tos_key.h"
 
 static GPtrArray *gdl_array=NULL;
 G_LOCK_DEFINE_STATIC(gdl_array);
 
 static void check_if_dl_complete(GDL_ENTRY *gdle);
 
+#define MAX_WAIT_TIME 420
+
 /************************************************************/
 /* scan gdl_array to find a GDL_ENTRY with the given gdl_id */
 /***************************************************************/
@@ -75,6 +82,37 @@ static int gdl_index(unsigned int gdl_id)
 	return -1;
 }
 
+/*************************************************************************/
+/* count the number of GDL_DL_ENTRY having is_running and is_started set */
+/*************************************************************************/
+/* input: a GPtrArray of GDL_DL_ENTRY */
+/* output: number of entry activated  */
+/**************************************/
+static int count_started_xfer(const GPtrArray *gdl_links)
+{
+	int nb=0;
+	int i;
+
+	if(gdl_links==NULL)
+		return nb;
+
+	for(i=0;i<gdl_links->len;i++)
+	{
+		GDL_DL_ENTRY *gdldle;
+
+		gdldle=g_ptr_array_index(gdl_links,i);
+		
+		if((gdldle!=NULL) &&
+			(gdldle->is_running) &&
+			(gdldle->is_started)
+		  )
+		{
+			nb++;
+		}
+	}
+	return nb;
+}
+
 /****************************************************************************/
 /* scan links of the given gdle to find a GDL_DL_ENTRY matching information */
 /****************************************************************************/
@@ -316,6 +354,28 @@ static GArray *split_free_planar_if_running(GArray *dl_ar, GArray **run_dl_ar, G
 	return dl_ar;
 }
 
+/***********************************************************/
+/* compute the lowest not yet downloaded position of a GDL */
+/***********************************************************/
+static unsigned long lowest_free_position(GDL_ENTRY *gdle)
+{
+	GArray *dl_ar;
+	unsigned long min_free_location=0;
+
+	dl_ar=ranges_to_planar(gdle,0);
+	if(dl_ar==NULL)
+		return min_free_location;			/* unable to compute ranges */
+	dl_ar=dl_planar_to_free_planar(dl_ar,gdle);
+
+	if(dl_ar->len>=1)
+	{
+		/* the first downloadable position in the file is the position of the first free segment */
+		min_free_location=(g_array_index(dl_ar,S_RNG,0)).b;
+	}
+	g_array_free(dl_ar,TRUE);
+	return min_free_location;	
+}
+
 /***************************************************/
 /* allocate a download range to the given DL_ENTRY */
 /***************************************************/
@@ -442,16 +502,22 @@ static int allocate_range(GDL_DL_ENTRY *gdldle, GDL_ENTRY *gdle)
 /*****************************/
 /* try to start its download */
 /*****************************/
-static void start_xdl(GDL_ENTRY *gdle, GDL_DL_ENTRY *gdldle)
+static void start_xdl(GDL_ENTRY *gdle, GDL_DL_ENTRY *gdldle, unsigned long int min_position)
 {
 	GString *nw;
 
 	gdldle->last_start_time=time(NULL);	/* always update time to avoid repeted try in short time */
 
+	if(gdldle->remote_fsize<=min_position)	/* the file of this source has no free range */
+		return;
+
 	if(!user_in_list(hub_user_list,gdldle->nickname->str))		/* user on the hub */
 	{
 		if((with_ddl==0)||(!check_uaddr_entry_by_name(gdldle->nickname->str)))	/* or DDL enable and user address is known */
-			return;												/* is required else, we leave */
+		{
+			if(with_dctclink==0)																	/* or inter DCTC communication is enabled */
+				return;																				/* is required else, we leave */
+		}
 	}
 
 	gdldle->is_running=1;
@@ -494,12 +560,18 @@ static void handle_one_gdl(GDL_ENTRY *gdle)
 {
 	int i;
 	time_t cur_time;
+	unsigned long int min_position;
 
 	if(gdle->gdl_links==NULL)
 		return;				/* nothing can be done */
 
+	if(count_started_xfer(gdle->gdl_links)>=max_running_source_per_gdl)
+		return;				/* enough source are running */
+
 	cur_time=time(NULL);
 
+	min_position=lowest_free_position(gdle);
+
 	/* try to start downloads on inactive links */
 	for(i=0;i<gdle->gdl_links->len;i++)
 	{
@@ -512,9 +584,10 @@ static void handle_one_gdl(GDL_ENTRY *gdle)
 		{
 			/* these few lines fix a bug somewhere else in the code */
 			/* sometimes, tasks are set to waiting but in fact, the real task (/XDL has been destroyed) */
-			if((ptr->is_started==0)&&((ptr->last_start_time+600)<cur_time))
+			if((ptr->is_started==0)&&((ptr->last_start_time+MAX_WAIT_TIME)<cur_time))
 			{
-				/* this XDL command has been sent more than 10 minutes and still not running ? BUG BUG BUG */
+				/* this XDL command has been sent more than 7 minutes and still not running ? */
+				/* this case can appear when dctc link is enabled and a try occurs */
 				ptr->is_started=0;
 				ptr->is_running=0;
 			}
@@ -527,7 +600,7 @@ static void handle_one_gdl(GDL_ENTRY *gdle)
 
 		if(!has_running_not_started_xdl(gdle,ptr))	/* to avoid a race and forever waiting transfer (which in fact no more exist) */
 																	/* we initiate only one connection at a time to the same user */
-			start_xdl(gdle,ptr);
+			start_xdl(gdle,ptr,min_position);
 	}
 }
 
@@ -574,6 +647,10 @@ static int append_this_range_entry_to_file(FILE *f, unsigned long *cur_pos, RANG
 		
 		/* read a segment slice */
 		want=MIN(sizeof(buf),to_append);
+
+		/* get 1 slice */
+		get_slice(bl_semid,GATHER_SEMA);
+
 		have=fread(buf,1,want,add);
 
 		if(have==-1)
@@ -794,6 +871,9 @@ static void do_gdl_autoscan(GDL_ENTRY *gdle)
 	if((gdle==NULL)||(gdle->autoscan==NULL)||(gdle->autoscan->len==0))
 		return;
 
+	if((gdle->gdl_links!=NULL) && (count_started_xfer(gdle->gdl_links)>=disable_gdl_as_when_enough_running) )
+		return;				/* enough source are running */
+
 	cur_time=time(NULL);
 	min_time=cur_time-auto_scan_delay;
 
@@ -813,7 +893,16 @@ static void do_gdl_autoscan(GDL_ENTRY *gdle)
 																					/* to reduce bandwidth waste, we ask a size of at least N-1 bytes */
 												gae->search_pattern);		/* it is a datatype?pattern in fact */
 			
-			add_new_sim_input(0,str->str);	/* yes, it is possible to sim_input DC command directly, not only dctc keyboard command, great isn't it ? :) */
+			if((main_sck!=-1)&&(cnx_in_progress==0))
+			{	/* if connected to the hub, send the message to this hub */
+				add_new_sim_input(0,str->str);	/* yes, it is possible to sim_input DC command directly, not only dctc keyboard command, great isn't it ? :) */
+			}
+
+			/* if dctc link is enable, also send the query to other hub */
+			send_dc_line_to_dctc_link_direct(NULL,str->str);
+
+			send_str_to_unode(str);		/* and also to UNODE */
+
 			g_string_free(str,TRUE);
 			gae->last_scan=cur_time;
 		}
@@ -844,6 +933,7 @@ static void	receive_and_decode_autoscan_result(GDL_ENTRY *gdle)
 			char *nick;
 			char *fname;
 			char *fsize;
+			char *ratio;
 			unsigned long file_size;
 
 			nick=buf+3;
@@ -857,6 +947,10 @@ static void	receive_and_decode_autoscan_result(GDL_ENTRY *gdle)
 				return;
 
 			*fname++='\0';
+
+			if(user_has_flag(nick,"IGNORE_SRCH"))		/* ignore this user search results ? */
+				return;
+
 			SKIP_SPACE(fname);
 			if(*fname=='\0')
 				return;
@@ -871,6 +965,12 @@ static void	receive_and_decode_autoscan_result(GDL_ENTRY *gdle)
 
 			file_size=strtoul(fsize,NULL,10);
 
+			ratio=strchr(fsize, ' ');
+			if(ratio==NULL)
+				return;
+			ratio++;
+
+
 			/* well, now, nick, fname and file_size are ready */
 			if(file_size==gdle->total_size)
 			{
@@ -883,6 +983,21 @@ static void	receive_and_decode_autoscan_result(GDL_ENTRY *gdle)
 				add_new_sim_input(0,new_src->str);
 				g_string_free(new_src,TRUE);
 			}
+
+			/* maybe we have a free slot here */
+			{
+				int ratio_free, ratio_ttl;
+
+				if(sscanf(ratio,"%d/%d",&ratio_free,&ratio_ttl)==2)
+				{
+					if((ratio_free>0)&&(with_sr_wake_up)&&(!tos_entry_exists(WAKEUGDL_TOSKEY,nick,strlen(nick),0)))
+					{  /* if there is free slot and option is enabled and this user waiting downloads have not been waked up since enough time, wake up them */
+						add_tos_entry(WAKEUGDL_TOSKEY,min_gdl_wake_up_delay,nick,strlen(nick),NULL,0);
+						send_dc_line_to_dctc_link(NULL,-1,"/WAKEUGDL",nick,NULL);
+						gdl_wake_up_sources(nick,1);		/* don't lock again or this will hang */
+					}
+				}
+   		}
 		}
 	}
 }
@@ -895,14 +1010,37 @@ static void gdl_thread(void *dummy)
 	int i;
 	fd_set rd;
 	int max_fd=-1;
+	time_t last_switch_time=0;
+	time_t cur_time;
 
 	while(1)
 	{
 		restart_loop:
 
+		cur_time=time(NULL);
+
 		FD_ZERO(&rd);
 		G_LOCK(gdl_array);
 
+		/* it is time to switch speed counter */
+		if((last_switch_time+10)<cur_time)
+		{
+			last_switch_time=cur_time;
+			i=0;
+			while(i<gdl_array->len)
+			{
+				GDL_ENTRY *gdle;
+				gdle=g_ptr_array_index(gdl_array,i);
+
+				if(gdle)
+				{
+					gdle->cur_spd=gdle->instant_spd;
+					gdle->instant_spd=0;
+				}
+				i++;
+			}
+		}
+
 		i=0;
 		while(i<gdl_array->len)
 		{
@@ -935,8 +1073,8 @@ static void gdl_thread(void *dummy)
 				}
 			}
 
-			if((main_sck!=-1)&&(cnx_in_progress==0)&&(!behind_fw))
-			{	/* if we are connected to the hub, we must handle GDL autoscan */
+			if(!behind_fw)
+			{	/* if we are connected to the hub or dctc link enabled, we must handle GDL autoscan */
 				do_gdl_autoscan(gdle);
 				
 				if(gdle->autoscan_sockfd!=-1)
@@ -1183,8 +1321,10 @@ static void update_gdl_cmd_of_this_gdle(GDL_ENTRY *gdle)
 /*******************************************/
 /* create a new GDL_ENTRY inside gdl_array */
 /*******************************************/
-/* output: 0= success, !=0: error */
-/**********************************/
+/* output: 0= success                  */
+/*         1=gdl with same name exists */
+/*         2=gdl with same idx exists  */
+/***************************************/
 int do_gdl_new(unsigned int gdl_id, char *local_filename, unsigned long int total_size)
 {
 	int ret=1;	/* default: error */
@@ -1220,6 +1360,8 @@ int do_gdl_new(unsigned int gdl_id, char *local_filename, unsigned long int tota
 			nw->total_size=total_size;
 			nw->gdl_links=g_ptr_array_new();
 			nw->dld_ranges=g_array_new(FALSE,FALSE,sizeof(RANGE_ENTRY));
+			nw->cur_spd=0;
+			nw->instant_spd=0;
 			nw->autoscan_sockfd=-1;
 			nw->autoscan=g_array_new(FALSE,FALSE,sizeof(GDL_AS_ENTRY));
 			create_dl_dir_and_lock(nw->local_fname->str,nw);
@@ -1232,9 +1374,11 @@ int do_gdl_new(unsigned int gdl_id, char *local_filename, unsigned long int tota
 	else
 	{
 		disp_msg(ERR_MSG,"do_gdl_new","a GDL with the same number still exists",NULL);
+		ret=2;
 	}
 	
 	G_UNLOCK(gdl_array);
+	SET_GSTATUS_GDL(gdl_array->len);
 	return ret;
 }
 
@@ -1358,6 +1502,35 @@ static int load_gdl_cmd(GDL_ENTRY *gdle, char *filename)
 			gae.gae_id=rand();
 			gae.search_pattern=g_strdup(fields[1]);
 			gdle->autoscan=g_array_append_val(gdle->autoscan,gae);
+
+			if(gdle->autoscan_sockfd==-1)
+			{
+				gdle->autoscan_sockfd=_x_udp(0);		/* create an UDP socket and found an unused port */
+				if(gdle->autoscan_sockfd==-1)
+				{
+					disp_msg(ERR_MSG,"load_gdl_cmd","WARNING: unable to create UDP socket for autoscan of this GDL","|lu",(unsigned long)(gdle->gdl_id),NULL);
+				}
+				else
+				{
+					/* search query requires local port number to receive search result */
+					struct sockaddr_in lcl;
+					int len_lcl=sizeof(lcl);
+
+					set_non_bloquant_sock(gdle->autoscan_sockfd);
+
+					if(getsockname(gdle->autoscan_sockfd,(void*)&lcl,&len_lcl)!=0)
+					{
+						disp_msg(ERR_MSG,"load_gdl_cmd","WARNING: unable to create UDP socket for autoscan of this GDL",
+																	"|lu",(unsigned long)(gdle->gdl_id),"error on getsockname",strerror(errno),NULL);
+						close(gdle->autoscan_sockfd);
+						gdle->autoscan_sockfd=-1;
+					}
+					else
+					{
+						gdle->autoscan_sock_port=ntohs(lcl.sin_port);
+					}
+				}
+			}
 		}
 		else
 				goto abrt;
@@ -1477,6 +1650,7 @@ void do_gdl_attach(char *filename)
 	}
 
 	G_UNLOCK(gdl_array);
+	SET_GSTATUS_GDL(gdl_array->len);
 }
 
 /***************************************************************************************************/
@@ -1535,6 +1709,57 @@ static void unlink_and_free_dl_entry(GDL_ENTRY *gdle, int gdl_links_index,int wi
 	}
 }
 
+/******************************************/
+/* Recursively delete the given directory */
+/******************************************/
+static void recur_del(GString *str)
+{
+	GString *name;
+	DIR *dir;
+	struct dirent *obj;
+
+	if((str==NULL)||(str->len==0))
+		return;
+
+	dir=opendir(str->str);
+	if(dir==NULL)
+	{
+		disp_msg(ERR_MSG,"recur_del","unable to open directory",str->str,NULL);
+		return;
+	}
+
+	name=g_string_new(NULL);
+	while((obj=readdir(dir))!=NULL)
+	{
+		if(!strcmp(obj->d_name,"."))
+			continue;
+		if(!strcmp(obj->d_name,".."))
+			continue;
+
+		g_string_sprintf(name,"%s/%s",str->str,obj->d_name);
+		if(unlink(name->str)==-1)
+		{
+			if(errno==EISDIR)
+				recur_del(name);
+		}
+	}
+
+	g_string_free(name,TRUE);
+	rmdir(str->str);
+}
+
+/**************************************************/
+/* delete the GDL directory having the given name */
+/**************************************************/
+static void discard_gdl_dir(GString *local_fname)
+{
+	GString *gdl_dir;
+
+	gdl_dir=g_string_new("");
+	g_string_sprintf(gdl_dir,"GDL/%s",local_fname->str);
+	recur_del(gdl_dir);
+	g_string_free(gdl_dir,TRUE);
+}
 
 /**********************************************************/
 /* detach a running GDL of this client                    */
@@ -1564,6 +1789,7 @@ int do_gdl_detach(unsigned int gdl_id)
 
 	gdle=g_ptr_array_remove_index_fast(gdl_array,idx);
 	G_UNLOCK(gdl_array);
+	SET_GSTATUS_GDL(gdl_array->len);
 
 	/* destroy all gdldle but without unlink files */
 	if(gdle->gdl_links!=NULL)
@@ -1618,16 +1844,19 @@ int do_gdl_detach(unsigned int gdl_id)
 		g_array_free(gdle->autoscan,TRUE);
 	}
 
-	if(gdle->local_fname)
-		g_string_free(gdle->local_fname,TRUE);
-
 	if(gdle->lock_fd!=-1)
 	{	/* release .lock file */
 		lockf(gdle->lock_fd,F_ULOCK,1);
 		close(gdle->lock_fd);
 	}
 		
+	if(gdle->local_fname)
+	{
+		g_string_free(gdle->local_fname,TRUE);
+	}
+
 	free(gdle);
+	return 0;
 }
 
 /*********************************************/
@@ -1782,6 +2011,7 @@ int do_gdl_end(unsigned int gdl_id,int override)
 	}
 	gdle=g_ptr_array_remove_index_fast(gdl_array,idx);
 	G_UNLOCK(gdl_array);
+	SET_GSTATUS_GDL(gdl_array->len);
 
 	/* destroy all gdldle */
 	if(gdle->gdl_links!=NULL)
@@ -1836,16 +2066,18 @@ int do_gdl_end(unsigned int gdl_id,int override)
 		g_array_free(gdle->autoscan,TRUE);
 	}
 
-	if(gdle->local_fname)
-		g_string_free(gdle->local_fname,TRUE);
-
 	if(gdle->lock_fd!=-1)
 	{	/* release .lock file */
 		lockf(gdle->lock_fd,F_ULOCK,1);
 		close(gdle->lock_fd);
 	}
 		
-	/* FIXME: GDL/fname is not destroyed in this function */
+	if(gdle->local_fname)
+	{
+		discard_gdl_dir(gdle->local_fname);
+		g_string_free(gdle->local_fname,TRUE);
+	}
+
 	free(gdle);
 
 	return 0;
@@ -2019,9 +2251,10 @@ static void check_if_dl_complete(GDL_ENTRY *gdle)
 /*        !=0 no range available        */
 /******************************************************************/
 /* if a range is assigned, *str is contains the $Get line to send */
-/* and *lfile contains the name of the local filename             */
+/* and *lfile contains the name of the local filename and         */
+/* *start_pos is the download start position (for information)    */
 /******************************************************************/
-int do_gdl_start(unsigned int gdl_id, char *nickname, pthread_t thread_id, GString **str,GString **lfile)
+int do_gdl_start(unsigned int gdl_id, char *nickname, pthread_t thread_id, GString **str,GString **lfile, unsigned long *start_pos)
 {
 	int ret=-1;
 	int i;
@@ -2060,6 +2293,7 @@ int do_gdl_start(unsigned int gdl_id, char *nickname, pthread_t thread_id, GStri
 									g_string_sprintf(*str,"$Get %s$%lu",gdldle->remote_fname->str,gdldle->range[0]+1);
 																						/* +1 because the 1st byte has the offset 1 instead of 0*/
 									*lfile=g_string_new(gdldle->temp_local_fname->str);
+									*start_pos=gdldle->range[0];
 									update_gdl_cmd_of_this_gdle(gdle);		/* reupdate the cmd */
 									break;
 								}
@@ -2249,7 +2483,14 @@ void do_gdl_fail(unsigned int gdl_id, char *nickname, char *local_fname, int is_
 {
 	int fnd=0;
 
-	disp_msg(DEBUG_MSG,"do_gdl_fail","|lu",gdl_id,nickname,local_fname,NULL);
+#if 0
+	if(local_fname!=NULL)
+#endif
+		disp_msg(DEBUG_MSG,"do_gdl_fail","|lu",gdl_id,nickname,local_fname,NULL);
+#if 0
+	else
+		disp_msg(INFO_MSG,"do_gdl_fail","|lu",gdl_id,nickname,"NULL fname",NULL);
+#endif
 
 	G_LOCK(gdl_array);
 	if(gdl_array!=NULL)
@@ -2267,35 +2508,75 @@ void do_gdl_fail(unsigned int gdl_id, char *nickname, char *local_fname, int is_
 				{
 					int j;
 
-					for(j=0;j<gdle->gdl_links->len;j++)
+#if 0
+					if(local_fname!=NULL)
 					{
-						GDL_DL_ENTRY *gdldle;
-
-						gdldle=g_ptr_array_index(gdle->gdl_links,j);
-
-						if(gdldle->is_started)
+#endif
+						for(j=0;j<gdle->gdl_links->len;j++)
 						{
-							if( (!strcmp(gdldle->nickname->str,nickname)) &&
-								 (!strcmp(gdldle->temp_local_fname->str,local_fname)) )
+							GDL_DL_ENTRY *gdldle;
+
+							gdldle=g_ptr_array_index(gdle->gdl_links,j);
+
+							if(gdldle->is_started)
 							{
-								/* ok, we have found GDL_DL_ENTRY */
-								fnd=1;
+								if( (!strcmp(gdldle->nickname->str,nickname)) &&												/* same nickname */
+								 	(!strcmp(gdldle->temp_local_fname->str,local_fname)))									/* and same fname */
+								{
+									/* ok, we have found GDL_DL_ENTRY */
+									fnd=1;
 
-								/* the download has at least successfully started, some data may have been downloaded */
-								gdl_post_download_update(gdle,gdldle);
-								gdldle->is_started=0;
-								g_string_free(gdldle->temp_local_fname,TRUE);
-								gdldle->temp_local_fname=NULL;
+									/* the download has at least successfully started, some data may have been downloaded */
+									gdl_post_download_update(gdle,gdldle);
+									gdldle->is_started=0;
+									g_string_free(gdldle->temp_local_fname,TRUE);
+									gdldle->temp_local_fname=NULL;
+	
+									gdldle->is_running=0;
+									gdldle->last_start_time=time(NULL);	/* to avoid immediat retry, adjust the start time */
+	
+									if(is_fatal)
+										unlink_and_free_dl_entry(gdle,j,0);
+									break;
+								}
+							}
+						}
+#if 0
+					}
+					else
+					{
+						/* this side is mainly used to fail a GDL entry when it has not yet started a download */
+						/* (==when the /XDL command fails) */
+						for(j=0;j<gdle->gdl_links->len;j++)
+						{
+							GDL_DL_ENTRY *gdldle;
 
-								gdldle->is_running=0;
-								gdldle->last_start_time=time(NULL);	/* to avoid immediat retry, adjust the start time */
+							gdldle=g_ptr_array_index(gdle->gdl_links,j);
 
-								if(is_fatal)
-									unlink_and_free_dl_entry(gdle,j,0);
-								break;
+							if((gdldle->is_running)&&(gdldle->is_started==0))		/* in try mode */
+							{
+								if(!strcmp(gdldle->nickname->str,nickname))												/* same nickname (no filename provided) */
+								{
+									/* ok, we have found GDL_DL_ENTRY */
+									fnd=1;
+
+									/* the download has at least successfully started, some data may have been downloaded */
+									gdl_post_download_update(gdle,gdldle);
+									gdldle->is_started=0;
+									g_string_free(gdldle->temp_local_fname,TRUE);
+									gdldle->temp_local_fname=NULL;
+	
+									gdldle->is_running=0;
+									gdldle->last_start_time=time(NULL);	/* to avoid immediat retry, adjust the start time */
+	
+									if(is_fatal)
+										unlink_and_free_dl_entry(gdle,j,0);
+									break;
+								}
 							}
 						}
 					}
+#endif
 				}
 				break;
 			}
@@ -2304,7 +2585,14 @@ void do_gdl_fail(unsigned int gdl_id, char *nickname, char *local_fname, int is_
 	G_UNLOCK(gdl_array);
 	if(!fnd)
 	{
-		disp_msg(ERR_MSG,"do_gdl_fail","|lu",gdl_id,nickname,local_fname,NULL);
+#if 0
+		if(local_fname!=NULL)
+#endif
+			disp_msg(ERR_MSG,"do_gdl_fail","|lu",gdl_id,nickname,local_fname,NULL);
+#if 0
+		else
+			disp_msg(INFO_MSG,"do_gdl_fail","|lu",gdl_id,nickname,"NULL fname",NULL);
+#endif
 	}
 }
 
@@ -2397,6 +2685,8 @@ unsigned int gdl_get_amount(unsigned int gdl_id, pthread_t ptid, unsigned long t
 							GArray *dl_ar;
 
 							/* ok, we have found GDL_DL_ENTRY */
+							gdle->instant_spd+= (total_pos-gdldle->cur_dled);
+
 							gdldle->cur_dled=total_pos;
 
 							/* compute the number of bytes we still can download */
@@ -2598,6 +2888,7 @@ void do_gdl_lst(int sck)
 													"|lu",(unsigned long)(gdle->dl_offset),		/* byte here before attachement */
 													"|lu",(unsigned long)(dld_size),					/* amount of bytes still here */
 													"|lu",(unsigned long)(gdle->start_time),		/* creation/attachement time */
+													"|lu",(unsigned long)(gdle->cur_spd),			/* number of bytes received in the previous 10 seconds range */
 													links_str->str,
 													range_str->str,
 													ascan_str->str,
@@ -2613,3 +2904,49 @@ void do_gdl_lst(int sck)
 	disp_msg(GDL_LST_END,"",NULL);
 }
 
+/*********************************************************************************************************/
+/* modify the timeout of waiting GDL sources having the given nickname to force them to retry immediatly */
+/*********************************************************************************************************/
+void gdl_wake_up_sources(char *nickname, int without_lock)
+{
+	int i;
+	time_t cur_time=time(NULL);
+
+	if(without_lock==0)
+		G_LOCK(gdl_array);
+	if(gdl_array!=NULL)
+	{
+		for(i=0;i<gdl_array->len;i++)
+		{
+			GDL_ENTRY *gdle;
+			
+			gdle=g_ptr_array_index(gdl_array,i);
+			if(gdle!=NULL)
+			{
+				GPtrArray *gdl_dl_entries;
+
+				gdl_dl_entries=gdle->gdl_links;
+				if(gdl_dl_entries!=NULL)
+				{
+					int j;
+
+					for(j=0;j<gdl_dl_entries->len;j++)
+					{
+						GDL_DL_ENTRY *ptr;
+
+						ptr=g_ptr_array_index(gdl_dl_entries,j);
+						if((ptr!=NULL)&&(ptr->is_running==0)&&(!strcmp(ptr->nickname->str,nickname)))
+						{
+							/* here is an entry in waiting status and having the wanted nickname */
+							/* try to start 1 second later */
+							ptr->last_start_time=cur_time-(MAX_WAIT_TIME-1);
+						}
+					}
+				}
+			}
+		}
+	}
+	if(without_lock==0)
+		G_UNLOCK(gdl_array);
+}
+
