@@ -38,6 +38,7 @@
 #include <errno.h>
 #include <getopt.h>
 #include <string.h>
+#include <db.h>
 #include <glib.h>
 #include <pthread.h>
 
@@ -51,7 +52,7 @@
 #include "macro.h"
 #include "dc_manage.h"
 #include "keyboard.h"
-#include "db.h"
+#include "mydb.h"
 #include "hl_locks.h"
 #include "main.h"
 #include "gts.h"
@@ -60,6 +61,8 @@
 #include "sema.h"
 #include "gdl.h"
 #include "uaddr.h"
+#include "status.h"
+#include "bdb.h"
 
 /*****************************************************/
 /* this mutex locks all pointers of user description */
@@ -72,6 +75,8 @@ HL_MUTEX user_info=HL_MUTEX_INIT;		/* this is a high level mutex allowing multip
 /* user description for DC hub */
 /*******************************/
 char *nickname=NULL;					/* user nickname */
+char *org_nickname=NULL;			/* user nickname (this is the one given on the command line. */
+											/* it is used when following hub redirection and reconnection) */
 char *user_desc=NULL;				/* user description (can be NULL) */
 char *cnx_type=NULL;					/* connection type. (ex: DSL,Cable,...) */
 char cnx_opt=0x1;                /* user status */
@@ -89,6 +94,9 @@ double offset_sizeof_data=0;     /* when the client has to return the size of da
                                  /* is added to the previous one */
 
 int free_dl_slot;						/* number of free download slot */
+											/* NOTE: this number can be negative for 2 reasons: */
+											/* 1) you have reduced the ttl_dl_slot value under the number of running uploads */
+											/* 2) a user having the IGNORE_SLOT_LIMIT flag is downloading */
 int ttl_dl_slot;						/* number of existing download slot */
 int dl_on;								/* this flag enables or disables the upload capability */
 											/* if ==1, other users can download from you */
@@ -120,6 +128,9 @@ GString *hubip=NULL;			/* IP of the hub where we are connected */
 unsigned int hub_port=411;	/* port of the hub where we are connected (411 is the default port) */
 GString *hubname=NULL;		/* Name of the hub where we are connected (filed when connected) */
 
+unsigned int unode_port=19284;	/* port number for UNODE, it is the current one */
+unsigned int wanted_unode_port=19284;	/* port number for UNODE. If you want to change UNODE port, change this value, not the one above */
+
 unsigned int recon_delay=30;	/* number of seconds before the hub connection loss and the /RECON */
 unsigned int auto_rebuild_delay=15*60;	/* number of seconds between 2 rebuilds of the shared database */
 unsigned int auto_scan_delay=10*60;	/* number of seconds between 2 GDL autoscan */
@@ -155,6 +166,28 @@ int having_emode=0;			/* depending of the connected hub, the current emode may d
 int with_md5sum=1;			/* enable/disable md5sum for content search */
 
 int with_ddl=0;				/* enable/disable dDL capability */
+int with_dctclink=0;			/* enable/disable link between dctc clients */
+int grab_ban_ip=0;			/* enable/disable IP grabbing of banned users */
+
+int with_dl_force=0;			/* enable/disable download forcing when a twin download occurs */
+
+int hide_absolute=1;			/* hide/show the / (\ seen on DC) at the beginning of the returned search result and file list */
+									/* default mode: hide absolute path */
+
+int abort_upload_when_user_leaves=0;
+									/* abort the remote user upload when a user leaves the hub */
+
+unsigned int max_running_source_per_gdl=10;	/* limit the number of running sources per GDL */
+unsigned int disable_gdl_as_when_enough_running=7;	/* disable one GDL's autoscan if enough running sources are available */
+
+int hide_kick=0;			/* hide/show the kick message displayed on the global chat when a user is kicked */
+
+int with_lazy_key_check=0;			/* enable/disable lazy $Key check */
+
+int with_incoming_wake_up=0;		/* enable/disable wakeup of waiting GDL source when a user with a "wanted" nickname enters the hub */
+int with_sr_wake_up=0;				/* enable/disable wakeup of waiting GDL source when a search result with a "wanted" nickname has free slot */
+
+int min_gdl_wake_up_delay=30;	/* at least 30 seconds between waking up source of the same user */
 
 /************************************************************/
 /* debug mode. ==0, no debug. !=0, debug mode               */
@@ -171,12 +204,16 @@ static int user_wants_to_quit=0;	/* this flag is changed when the user hit ctrl-
 /*******************************************************************/
 GString *local_dctc_sock_path=NULL;					/* local name */
 int local_sck=-1;											/* main local socket (to accept connection) */
+GString *local_dctc_udp_sock_path=NULL;			/* local name */
+int local_sck_udp=-1;									/* main local UDP socket (to receive command without being connected (inter DCTC link)) */
 
 GArray *local_client_socket=NULL;					/* array of all locally connected socket */
 G_LOCK_DEFINE(local_client_socket);
 
 GString *dctc_dir=NULL;									/* == $(HOME)/.dctc */
 
+GString *dctc_active_client_file=NULL;
+
 const char *dc_version="1,0091";								/* DC version to use */
 
 /*****************************************/
@@ -228,8 +265,15 @@ void hub_disconnect(HUBDISC_FLAG exit_flag)
 			main_sck=-1;
 		}
 
+		set_client_status(IS_OFFLINE);
 		reset_hub_user_list();		/* else /ULIST will return the old user list */
 		display_cnx_status();
+		reset_last_myinfo();
+
+		/* reset the nickname to its original value */
+		if(nickname!=NULL)
+			free(nickname);
+		nickname=strdup(org_nickname);
 
 		add_new_sim_input(recon_delay,"/RECON");
 		add_new_sim_input(0,"/VARS");
@@ -281,9 +325,16 @@ void hub_disconnect(HUBDISC_FLAG exit_flag)
 	}
 
 	/* and end */
+	del_client_status();
 	disp_msg(INFO_MSG,"hub_disconnect","end",NULL);
 	if(local_dctc_sock_path!=NULL)
 		unlink(local_dctc_sock_path->str);
+	if(local_dctc_udp_sock_path!=NULL)
+		unlink(local_dctc_udp_sock_path->str);
+
+	/* exit BerkeleyDB lib */
+	do_berkeley_exit();
+
 	sleep(1);		/* without this sleep, some of the latest disp_msg doesn't work */
 	exit(1);
 }
@@ -315,6 +366,7 @@ static void	expire_old_xfer(void)
 				/* remove action from the list */
 				g_ptr_array_remove_index(waiting_revcon,i);
 
+#if 0
 				if(!strncmp(ptr->action_to_do->str,"DL/",3))
 				{
 					/* we will requeue this xfer inside sim_input */
@@ -325,6 +377,8 @@ static void	expire_old_xfer(void)
 						char *t;
 						char sep;
 
+						uaddr_increase_error_flag(ptr->remote_nick->str);
+
 						sep=ptr->action_to_do->str[sizeof("DL/")-1]; /* get the separator */
 
 						local=g_string_new(ptr->action_to_do->str + sizeof("DL/") -1 + 1 /* ignore the separator */ );
@@ -364,8 +418,12 @@ static void	expire_old_xfer(void)
 					}
 					disp_msg(XFER_DL_UNQUEUED,NULL, ptr->remote_nick->str,ptr->action_to_do->str,NULL);
 				}
-				else if(!strncmp(ptr->action_to_do->str,"LS/",3))
+				else 
+#endif
+				if(!strncmp(ptr->action_to_do->str,"LS/",3))
 				{
+					uaddr_increase_error_flag(ptr->remote_nick->str);
+
 					/* we will requeue this xfer inside sim_input */
 					{
 						GString *sim;
@@ -385,8 +443,11 @@ static void	expire_old_xfer(void)
 				}
 				else if(!strncmp(ptr->action_to_do->str,"XDL|",4))
 				{
-					/* notify to GDL the transfer never start */
 					gchar **fields;
+
+					uaddr_increase_error_flag(ptr->remote_nick->str);
+
+					/* notify to GDL the transfer never start */
 					fields=g_strsplit(ptr->action_to_do->str,"|",3);		/* there is always 3 fields, never less, never more */
 					do_gdl_abort(strtoul(fields[1],NULL,10),fields[2]);
 					g_strfreev(fields);
@@ -568,6 +629,46 @@ static void	manage_unix_socket(int main_sck, int sck, int num)
 	g_string_free(s,TRUE);
 }
 
+/******************************************************************/
+/* read data from UDP socket and send it to the keyboard function */
+/******************************************************************/
+static void	manage_local_udp_socket(int sck)
+{
+	char buf[8192];
+	int ret;
+
+	if(sck==-1)	 /* this should never happen but who knows .... */
+		return;
+
+	ret=recv(sck,buf,sizeof(buf)-1,MSG_NOSIGNAL);
+	if(ret==-1)
+	{
+		disp_msg(ERR_MSG,"manage_local_udp_socket","recv error",strerror(errno),NULL);
+		return;
+	}
+	if(ret<2)
+	{
+		return;
+	}
+
+	if(with_dctclink==0)
+		return;
+
+	buf[ret]='\0';
+	printf("unix_udp_socket: %s\n",buf);
+
+	/* 2 small tests: a valid command always starts with '$' or '~' (conditionnal run) or '*' output messages */
+	/*                or '[' (a DCTC output message to relay] */
+	/*                it can also be a DCTC command (not used by DCTC itself but can be by UI) */
+	/*                and always ends with a '\n' */
+	if((buf[0]!='$')&&(buf[0]!='~')&&(buf[0]!='*')&&(buf[0]!='/')&&(buf[0]!='['))
+		return;
+	if(buf[ret-1]!='\n')
+		return;
+
+	keyboard_input(main_sck,buf);
+}
+
 /*******************************************************************/
 /* this function is called when the hub socket is ready to be used */
 /* (2nd stage of the connection). We can have 2 cases:             */
@@ -714,6 +815,12 @@ static void main_loop()
 			n=max(n,local_sck);
 		}
 
+		if(local_sck_udp!=-1)
+		{
+			FD_SET(local_sck_udp,&rd);
+			n=max(n,local_sck_udp);
+		}
+
 		/* add unix client socket */
 		G_LOCK(local_client_socket);
 		for(i=0;i<local_client_socket->len;i++)
@@ -760,6 +867,11 @@ static void main_loop()
 				/* someone tries to reach the client from unix socket */
 				manage_local_socket(local_sck);
 			}
+			else if((local_sck_udp!=-1)&&(FD_ISSET(local_sck_udp,&rd)))
+			{
+				/* someone tries to reach the client from unix socket */
+				manage_local_udp_socket(local_sck_udp);
+			}
 			else
 			{
 				/* check unix client socket */
@@ -898,8 +1010,8 @@ static void display_usage(char *fname)
 						"                                  (default: 412 (like Direct Connect)). Be\n"
 						"                                  careful, value below 1024 is only allowed to\n"
 						"                                  privileged users (root)\n"
-						"  -g, --hubip=HUBIP[:PORT]        IP of the hub to connect to. If no port is given\n"
-						"                                  the default port (411) is used\n"
+						"  -g, --hubip=HUBIP[:PORT]        IP of the hub to connect to. If no port is\n"
+						"                                  given the default port (411) is used\n"
 						"  -f, --firewall                  enable passive mode because host is behind a\n"
 						"                                  firewall (default: not behind a firewall)\n"
 						"  -x, --no_xfer                   Remote users cannot download (same as /DLOFF)\n"
@@ -914,12 +1026,17 @@ static void display_usage(char *fname)
 						"                                  a hub refusing you just because you don't have\n"
 						"                                  the latest version of DC. This DCTC supports\n"
 						"                                  DC protocol version %s\n"
-						"  -u, --upload=NUM                Upload bandwidth limit. NUM is the number of\n"
+						"  -u, --upload=NUM1,NUM2,NUM3     Bandwidth limit. NUM1 is the number of\n"
 						"                                  512 bytes per second allowed to be sent\n"
+						"                                  NUM2 is the number of 512 bytes per second\n"
+						"                                  allowed to be received. NUM3 is the number\n"
+						"                                  of 8KBytes bloc per second that can be copied\n"
+						"                                  when client gathers GDLs parts.\n"
 						"                                  (default: unlimited [in fact, it is 16MB/s])\n"
 						"                                  Note: this flag is only taken into account\n"
 						"                                  if this client creates the bandwidth limit\n"
-						"                                  semaphore, use /UBL to change the value later\n"
+						"                                  semaphore, use /[UDG]BL to change the value\n"
+						"                                  later\n"
 						"  -m, --mdc=NUM                   enable non DC hub compatible command\n"
 						"                                  this flag is for future usage. Current default\n"
 						"                                  value is 0.\n"
@@ -933,6 +1050,17 @@ static void display_usage(char *fname)
 						"  -5, --nomd5sum                  Do not compute MD5 sum for each shared file\n"
 						"                                  This sum is used for content search.\n"
 						"                                  Default: enabled\n"
+						"  -k, --dctclink                  some commands can be received from or send to\n"
+						"                                  other running dctc clients. This is useful to\n"
+						"                                  improve search and download efficiency.\n"
+						"  -S, --socksaddr=ADDR            address (IP or name) of the SOCKS proxy to use\n"
+                  "                                  Default: no SOCKS proxy is used\n"
+                  "  -P, --socksport=NUM             port of the SOCKS proxy. Default: 1080\n"
+                  "                                  Only used if -S used\n"
+                  "  -U, --socksuid=STRING           If the SOCKS proxy requires a userID, this is\n"
+						"                                  it. Default: none\n"
+						"  -D, --virtual_share_dir=DIR     Define a directory contains files describing\n"
+						"                                  virtually shared files. (see Docs/Vshare).\n"
 						"\n"
 						"Be careful, most of the information you provide can't contain the following\n"
 						"characters because Direct Connect uses them internally: | $\n"
@@ -1057,7 +1185,6 @@ static void start_rebuild_db_thread(void)
 	}
 }
 
-
 /* beginning of the code */
 int main(int argc,char **argv)
 {
@@ -1082,15 +1209,22 @@ int main(int argc,char **argv)
 												{"upload",required_argument,NULL,'u'},		/* upload bandwidth limit */
 												{"mdc",required_argument,NULL,'m'},			/* enable non compatible DC command */
 												{"precmd",required_argument,NULL,'b'},		/* send a command to execute before hub connection */
-												{"nomd5sum",no_argument,NULL,'5'},	/* don't compute md5 sum for each file of the database */
+												{"nomd5sum",no_argument,NULL,'5'},			/* don't compute md5 sum for each file of the database */
+												{"dctclink",no_argument,NULL,'k'},			/* enable inter dctc communication */
+												{"socksaddr",required_argument,NULL,'S'},	/* set the address of the socks proxy to use */
+												{"socksport",required_argument,NULL,'P'},	/* set the port of the socks proxy to use */
+												{"socksuid",required_argument,NULL,'U'},	/* set the userID to use with the socks proxy */
+												{"virtual_share_dir",required_argument,NULL,'D'},	/* directory containing files containing virtual shared files */
 												{NULL,0,NULL,'\0'}					/* last option */
 											};
-	static const char *short_opt="hn:i:c:e:d:s:o:a:p:g:fxwtlv:u:m:b:5";
+	static const char *short_opt="hn:i:c:e:d:s:o:a:p:g:fxwtlv:u:m:b:5kS:P:U:D:";
 										
 	int ch;
 	int detach_from_tty=0;
 	int wait_lnk=0;
 	int spd_limit=SEMVMX;
+	int dl_spd_limit=SEMVMX;
+	int gath_spd_limit=SEMVMX;
 
 	/* the 2 sh_* are used to stored -s directories */
 	GStringChunk *sh_gsc;
@@ -1100,6 +1234,8 @@ int main(int argc,char **argv)
 	GStringChunk *pre_gsc;
 	GPtrArray *pre_gpa;
 
+	char *virtual_share_path=NULL;
+
 	disp_msg(INFO_MSG,NULL,"Direct Connect Text Client v" VERSION ,NULL);
 
 	if(argc==1)
@@ -1108,6 +1244,9 @@ int main(int argc,char **argv)
 		exit(0);
 	}
 
+	/* initialize random number generator else all clients will generate the same digit at the end of the nick */
+	srand(time(NULL));
+
 	/* set default values */
 	nickname=strdup("Noname");
 	user_desc=NULL;
@@ -1145,7 +1284,7 @@ int main(int argc,char **argv)
 			case 'i':	if(user_desc!=NULL)
 								free(user_desc);
 							user_desc=strdup(optarg);
-							check_string(nickname);
+							check_string(user_desc);
 							break;
 
 			case 'c':	if(cnx_type!=NULL)
@@ -1272,9 +1411,13 @@ int main(int argc,char **argv)
 			case 'v':	dc_version=strdup(optarg);
 							break;
 
-			case 'u':	spd_limit=atoi(optarg);
+			case 'u':	sscanf(optarg,"%d,%d,%d",&spd_limit,&dl_spd_limit,&gath_spd_limit);
 							if((spd_limit<0)||(spd_limit>SEMVMX))
 								spd_limit=SEMVMX;
+							if((dl_spd_limit<0)||(dl_spd_limit>SEMVMX))
+								dl_spd_limit=SEMVMX;
+							if((gath_spd_limit<0)||(gath_spd_limit>SEMVMX))
+								gath_spd_limit=SEMVMX;
 							break;
 
 			case 'b':	g_ptr_array_add(pre_gpa,g_string_chunk_insert(pre_gsc,optarg));
@@ -1283,12 +1426,48 @@ int main(int argc,char **argv)
 			case '5':	with_md5sum=0;
 							break;
 
+			case 'k':	with_dctclink=1;
+							break;
+
+			case 'S':	/* SOCKS address */
+							if(socks_ip!=NULL)
+								free(socks_ip);
+							if(strlen(optarg)==0)
+								socks_ip=NULL;
+							else
+								socks_ip=strdup(optarg);
+							break;
+
+			case 'P':	/* SOCKS port */
+							socks_port=strtoul(optarg,NULL,10);
+							if((socks_port==0)||(socks_port>65534))
+							{
+								socks_port=1080;
+								disp_msg(ERR_MSG,NULL,"SOCKS port is invalid. Must be >0 and <65535, have:",optarg,"SOCKS port forced to 1080",NULL);
+							}
+							break;
+
+			case 'U':	/* SOCKS user ID */
+							if(socks_name!=NULL)
+								free(socks_name);
+							socks_name=strdup(optarg);
+							break;
+
+			case 'D':	virtual_share_path=strdup(optarg);
+							break;
+
 			default:
 							fprintf(stderr,"Unknown option: %c\n",ch);
 							exit(1);
 		}
 	}
 
+	org_nickname=strdup(nickname);
+
+	/* if SOCKS is used, force passive mode */
+	if(socks_ip!=NULL)
+		behind_fw=1;
+
 	/* at the beginning, all download slots are free */
 	free_dl_slot=ttl_dl_slot;
 
@@ -1422,6 +1601,14 @@ int main(int argc,char **argv)
 				exit(1);
 			}
 		}
+		else
+		{
+			if(access(local_dctc_sock_path->str,R_OK|W_OK|X_OK))
+			{
+				disp_msg(ERR_MSG,"","access to ",local_dctc_sock_path->str," ",strerror(errno),NULL);
+				exit(1);
+			}
+		}
 
 		dctc_dir=g_string_new(local_dctc_sock_path->str);
 
@@ -1434,9 +1621,19 @@ int main(int argc,char **argv)
 				exit(1);
 			}
 		}
+		else
+		{
+			if(access(local_dctc_sock_path->str,R_OK|W_OK|X_OK))
+			{
+				disp_msg(ERR_MSG,"","access to ",local_dctc_sock_path->str," ",strerror(errno),NULL);
+				exit(1);
+			}
+		}
+
 
 		g_string_sprintfa(local_dctc_sock_path,"/dctc-%08X-%s",getpid(),org_hubip->str);
 
+		/* create local TCP socket (dctc<=>ui com) */
 		local_sck=socket(AF_UNIX,SOCK_STREAM,0);
 		if(local_sck==-1)
 		{
@@ -1458,12 +1655,37 @@ int main(int argc,char **argv)
 			listen(local_sck,3);
 		}
 
+		/* create local UDP socket (dctc<=>dctc com) */
+		local_dctc_udp_sock_path=g_string_new(local_dctc_sock_path->str);
+		local_dctc_udp_sock_path=g_string_append(local_dctc_udp_sock_path,".udp");
+
+		local_sck_udp=socket(AF_UNIX,SOCK_DGRAM,0);
+		if(local_sck_udp==-1)
+		{
+			perror("local_sck_udp - socket");
+			exit(1);
+		}
+		
+		{
+			struct sockaddr_un name;
+
+			name.sun_family=AF_UNIX;
+			strcpy(name.sun_path,local_dctc_udp_sock_path->str);
+			if(bind(local_sck_udp,(void *)&name,sizeof(struct sockaddr_un)))
+			{
+				perror("local_sck_udp - bind");
+				exit(1);
+			}
+		}
+
+		/* if necessary, wait for someone to connect on the TCP socket */
 		if(wait_lnk)
 		{
 			manage_local_socket(local_sck);
 		}
 	}
 
+
 	last_cmd_time=time(NULL);
 	display_cnx_status();
 
@@ -1479,19 +1701,32 @@ int main(int argc,char **argv)
 		bl_fname=g_string_new(dctc_dir->str);
 		bl_fname=g_string_append_c(bl_fname,'/');
 		bl_fname=g_string_append(bl_fname,unm.nodename);
-		if(do_sema_init(bl_fname->str,&bl_semid,spd_limit))
+		g_string_sprintfa(bl_fname,".%u",SEMA_ARRAY_LEN);
+		if(do_sema_init(bl_fname->str,&bl_semid,spd_limit, dl_spd_limit, gath_spd_limit))
 		{
 			fprintf(stderr,"Unable to initialize Bandwidth limitation semaphore.\n");
 			exit(1);
 		}
 	}
 
+	/* initialize BerkeleyDB lib */
+	do_berkeley_init();
+
+	dctc_active_client_file=g_string_new(dctc_dir->str);
+	dctc_active_client_file=g_string_append(dctc_active_client_file,"/gstatus");
+
 	/* its time to add the shared directories to the list of shared directories */
 	{
 		int i;
 		float cur_pos=0;
 		float step=100.0/(sh_gpa->len);
 
+		if(virtual_share_path!=NULL)
+		{
+			set_vshare_directory(virtual_share_path);
+			free(virtual_share_path);
+		}
+
 		disp_msg(PROGRESS_BAR,"","init_share","0","Initialize shared file database",NULL);
 		
 		for(i=0;i<sh_gpa->len;i++)
@@ -1515,15 +1750,20 @@ int main(int argc,char **argv)
 		g_string_chunk_free(sh_gsc);
 	}
 
+	set_client_status(IS_ONLINE);
 	start_rebuild_db_thread();
 
 	start_gdl_thread();
 
 	main_loop();
 
+	/* despite of what it looks like, everything below this comment is never used */
 	exit_gts();
 	exit_uaddr();
 
+	/* exit BerkeleyDB lib */
+	do_berkeley_exit();
+
 	/* close internet socket */
 	close(main_sck);
 	main_sck=-1;
@@ -1531,6 +1771,8 @@ int main(int argc,char **argv)
 	/* close unix sockets */
 	close(local_sck);
 	local_sck=-1;
+	close(local_sck_udp);
+	local_sck_udp=-1;
 
 	G_LOCK(local_client_socket);
 	while(local_client_socket->len!=0)
@@ -1545,6 +1787,7 @@ int main(int argc,char **argv)
 	G_UNLOCK(local_client_socket);
 	
 	unlink(local_dctc_sock_path->str);
+	unlink(local_dctc_udp_sock_path->str);
 
 	disp_msg(INFO_MSG,"main","end of client",NULL);
 	return 0;
